package edu.utexas.nlp;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.*;

public class MainForInstructors {

  static Random r = new Random();

  public static List<List<String>> readWikitext(String path, int maxLines) {
    System.out.println("Started reading from file " + path);
    List<List<String>> lines = new ArrayList<List<String>>();
    BufferedReader reader;
    try {
      reader = new BufferedReader(new FileReader(path));
      String line = reader.readLine();
      int numLines = 0;
      while (line != null && (maxLines < 0 || numLines < maxLines)) {
        if (line.trim().length() > 0) {
          List<String> thisLine = new ArrayList<>();
          thisLine.add(BigramLanguageModel.BEGIN_SYMBOL);
          String[] split = line.split(" ");
          for (String word : split) {
            if (word.trim().length() > 0) {
              thisLine.add(word);
            }
          }
          thisLine.add(BigramLanguageModel.END_SYMBOL);
          lines.add(thisLine);
          numLines += 1;
        }
        line = reader.readLine();
      }
      reader.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
    System.out.println("Read " + lines.size() + " lines");
    return lines;
  }

  /**
   * Checks the normalization of lm in a few different contexts
   * @param lm
   */
  public static void checkNormalization(BigramLanguageModel lm) {
    checkNormalization(lm, "the");
    checkNormalization(lm, "asked");
    checkNormalization(lm, "did");
  }

  /**
   * Checks the normalization of the LM in the given prevWord context. Prints a warning message if the sum
   * of the probabilities of possible next words is significantly different from 1.
   * @param lm
   * @param prevWord
   */
  public static void checkNormalization(BigramLanguageModel lm, String prevWord) {
    double totalProb = 0.0;
    for (String word : lm.getVocabulary()) {
      totalProb += lm.getProbability(prevWord, word);
    }
    if (Math.abs(totalProb - 1.0) > 1e-3) {
      System.out.println("ERROR: normalization test failed: probabilities sum to " + totalProb + " rather than 1");
    } else {
      System.out.println("Okay! Sums to " + totalProb + " after context \"" + prevWord + "\"");
    }
  }

  /**
   * Prints the top 5 most likely next words after prevWord according to the lm
   * @param lm
   * @param prevWord
   */
  public static void queryLm(BigramLanguageModel lm, String prevWords) {
    String prevWord;
    if (prevWords.contains(" ")) {
      prevWord = prevWords.substring(prevWords.lastIndexOf(' ') + 1);
    } else {
      prevWord = prevWords;
    }
    Map<String,Double> nextStepMap = new HashMap<>();
    for (String word: lm.getVocabulary()) {
      nextStepMap.put(word, lm.getProbability(prevWord, word));
    }
    List<Map.Entry<String,Double>> probsNextStep = new ArrayList<>();
    probsNextStep.addAll(nextStepMap.entrySet());
    Collections.sort(probsNextStep, new Comparator<Map.Entry<String, Double>>() {
      public int compare(Map.Entry<String, Double> o1, Map.Entry<String, Double> o2) {
        return o2.getValue().compareTo(o1.getValue());
      }
    });
    String result = "Top 5 words and probabilities after \"" + prevWord + "\": ";
    for (int i = 0; i < Math.min(5, probsNextStep.size()); i++) {
      result += "(" + probsNextStep.get(i).getKey() + ": " + probsNextStep.get(i).getValue() + "), ";
    }
    System.out.println(result.substring(0, result.length() - 2));
  }

  /**
   * @param lm
   * @param contextWord An initial word to seed the sentence with
   * @return A randomly-sampled word to follow contextWord according to the probabilities from lm.getProbability.
   *     Hint: you'll want to use something like
   *     Random r = new Random();
   *     r.nextDouble();
   *     to get a random number, then follow the scheme described in the video for how to turn that random number
   *     into a random word.
   */
  public static String sampleWord(BigramLanguageModel lm, String contextWord) {
    double sample = r.nextDouble();
    double probSum = 0.0;
    String latestWord = "";
    for (String word : lm.getVocabulary()) {
      latestWord = word;
      double prob = lm.getProbability(contextWord, word);
      probSum += prob;
      if (probSum > sample) {
        return word;
      }
    }
    return latestWord;
  }

  /**
   * @param lm
   * @param contextWord An initial word to seed the sentence with
   * @return Up to 10 words as a continuation of context_word by repeatedly sampling the next word
   */
  public static List<String> sampleSentence(BigramLanguageModel lm, String contextWord) {
    List<String> sent = new ArrayList<String>();
    sent.add(contextWord);
    for (int i = 0; i < 10; i++) {
      String nextWord = sampleWord(lm, contextWord);
      contextWord = nextWord;
      sent.add(nextWord);
      if (nextWord.equals(BigramLanguageModel.END_SYMBOL)) {
        return sent;
      }
    }
    return sent;
  }

  public static void evaluateLm(BigramLanguageModel lm, List<List<String>> data) {
    double totalLogProb = 0.0;
    int wordCounts = 0;
    Collection<String> vocab = lm.getVocabulary();
    int ignoredWords = 0;
    for (List<String> sent : data) {
      // Score everything from 1 until the end, ignoring BOS
      for (int i = 1; i < sent.size(); i++) {
        if (vocab.contains(sent.get(i))) {
          totalLogProb += -Math.log(lm.getProbability(sent.get(i - 1), sent.get(i)));
          wordCounts += 1;
        } else {
          ignoredWords += 1;
        }
      }
    }
    System.out.println("Perplexity: " + Math.exp(totalLogProb/wordCounts) + " with " + ignoredWords + " words ignored");
  }

  public static void main(String[] args) {
    List<List<String>> trainLines = readWikitext("wiki.train.tokens", -1);
    List<List<String>> validLines = readWikitext("wiki.valid.tokens", -1);
    BigramLanguageModel lm = BigramLanguageModel.estimate(trainLines);
    checkNormalization(lm);
    queryLm(lm, "the");
    queryLm(lm, "like");
    queryLm(lm, "to");
    queryLm(lm, "I like to");
    queryLm(lm, "I want to");
    // Draws 10 samples of words that follow "I"
    for (int i = 0; i < 10; i++) {
      System.out.println(sampleWord(lm, "I"));
    }
    // Draws 10 samples of sentences starting with "I"
    for (int i = 0; i < 10; i++) {
      System.out.println(sampleSentence(lm, "I"));
    }
    // Draws 10 samples of sentences starting with "It"
    for (int i = 0; i < 10; i++) {
      System.out.println(sampleSentence(lm, "It"));
    }
  }
}

